#!/home/lua
mime = {};
require('url')
require('web.auth')
require('lp') -- loading perax.lp

-- $Id: httpd.lua,v 1.9 2008-10-08 12:57:27 obarthel Exp $

global_client = 0
global_path = '/home/lua-perax/web/'
global_method_list = {}
global_cookie = 0
global_session = nil
listener = 0

mem = perax:open_mem()
com = perax:open_com()
dm = perax:open_dm()
pio = perax:open_pio()

function out (s) 
	libhttpd.write(global_client, s)
end

function load_module(name) 
	r, vi = pcall(require, 'web.' .. name)
	if r ~= true then
		print (vi)
		sendError( client, 501, "Error loading module module "  ..vi )
		return nil
	end
	return 0
end

function get_method (path) 
	_, _, name =  string.find(path, '/(.[^/]*)')

	if name == nil then
		name =  string.sub(path, 2)
	end

	_, _, _, extension = string.find(path, '/(.[^/]*)/(.[^/]*)')

	if path == '/' then
		extension = 'index'
		name = 'index'
	end

	if extension == nil then
		extension = 'index'
	else
		-- print ("Extension found " .. extension)
	end

	if name == '' then 
		name = 'index'
	end

	file = global_path .. name .. '.lua'

	if string.find(name, '\.js') then
		file_fixed = global_path .. 'js/' .. name  
	elseif string.find(name, '\.html') then
		file_fixed = global_path .. 'html/' .. name 
	elseif string.find(name, '\.css') then
		file_fixed = global_path .. 'css/' .. name 
	elseif string.find(name, '\.jpg') or string.find(name, '\.png') or string.find(name, '\.ico') then
		file_fixed = global_path .. 'images/' .. name 
	elseif string.find(name, '\.lp') then
		file_fixed = global_path .. 'templates/' .. name 
	else
		file_fixed = "__"
	end
		
	r, val = pcall(lfs.attributes, file)
	-- print (file .. '   result : ' .. tostring(val))
	
	if val == nil then
		--print (file_fixed)
		--f = io.open(file_fixed)
		
		if string.find(name, '\.lp$') then
			f = io.open(file_fixed)
			content = f:read('*all')
			inter = perax.lp.translate(content)
			r, val = pcall(loadstring, inter)
			if r == true then
				return val
			else
				return nil
			end
		end
		
		if file_fixed ~= nil then
			return function (p) 
			--	out(f:read('*all'))
			--	f:close()
			    libhttpd.sendfile(global_client, file_fixed)
			end
		else 
			sendError( client, 501, "Error loading module ")
			return nil
		end

	end

	if global_method_list[name] == nil then
		if load_module(name) ~= nil then
			-- print("Loading Module " .. name)
			global_method_list[name] = val.change
		else 
			return nil
		end

	elseif val.change ~= global_method_list[name] then
		-- print ('reloading module '  .. name)
		package.loaded['web.' .. name] = nil
		load_module(name)
		global_method_list[name] = val.change
	end

	return _G[name .. "_" .. extension]

end

function start_server( port )
	running = 1;

	print ('starting server')
	listener = libhttpd.bind( port );

	if ( listener == nil ) then
		error("Error in bind()");
	end

	print ('Listen on port ' .. port)

	while ( running == 1 ) do
		r, v = pcall(processConnection, listener);
		if r ~= true then
			print "error Processing the connection"
		end
	end

	libhttpd.close( listener );
end




--
--  Process a single incoming connection.
--
function processConnection( listener ) 
	--
	--  Accept a new connection
	--
	client,ip = libhttpd.accept( listener );
	global_client = client

	found   = 0;  -- Found the end of the HTTP headers?
	chunk   = 0;  -- Count of data read from client libhttpd
	size    = 0;  -- Total size of incoming request.
	code    = 0;  -- Status code we send to the client
	request = ""; -- Request body read from client.


	--
	--  Read in a response from the client, terminating at the first
	-- '\r\n\r\n' line - this is the end of the HTTP header request.
	--
	--  Also break out of this loop if we read ten packets of data
	-- from the client but didn't manage to find a HTTP header end.
	-- This should help protect us from DOSes.
	--
	while ( ( found == 0 ) and ( chunk < 10 ) ) do
		length, data = libhttpd.read(client);
		if ( length < 1 ) then
			found = 1;
		end

		size    = size + length;
		request = request .. data;

		-- print (request)
		position,len = string.find( request, '\r\n\r\n' );
		if ( position ~= nil )  then
			found = 1;
		end

		chunk = chunk + 1;
	end

	--print (request)
	_, _, method, path, major, minor  = string.find(request, "([A-Z]+) (.+) HTTP/(%d).(%d)");

	if ( method ~= "GET" ) then
		error = "Method not implemented";

		if ( method == nil ) then
			error = error .. ".";
		else
			error = error .. ": "  -- .. urlEncode( method );
		end

		size = sendError( client, 501, error );
		libhttpd.close( client );
		return size, "501";
	end


	path = urlDecode( path );

	_, _, host    = string.find(request, "Host: ([^:\r\n]+)");
	_, _, agent   = string.find(request, "Agent: ([^\r\n]+)");

	-- user gestion
	_, _, global_cookie = string.find(request, "Cookie:.*luap4=([^;\r\n]+)");
	if global_cookie then
		global_session = auth_session_list[global_cookie]
	else
		global_session = nil
	end

	_, _, referer = string.find(request, "Referer: ([^\r\n]+)");


	size, code = handleRequest(  host, path, client );

	if ( agent == nil )   then agent   = "-" end;
	if ( referer == nil ) then referer = "-" end;
	if ( code == nill )   then code    = 0 ; end;

	-- pcall(logAccess,  host, ip, path, code, size, agent, referer, major, minor );

	if ( running == 0 ) then
		print( "Terminating as per request from " .. ip );
	end

	libhttpd.close( client );
end



--
--  Attempt to serve the given path to the given client
--
function handleRequest(  host, path, client )

	query = parse(path)   

	parameters = {}

	if query.query ~= nil then 
		string.gsub(query.query,'\?*(.[^=?]*)=(.[^&]*)&*',function(v, value) parameters[v] = value end )
	end

	m = get_method(query.path)

	if m ~= nil then
		if string.find(query.path, '\.jpg') then
			mtype  = 'image/jpg'
		elseif  string.find(query.path, '\.png') then
			-- print(query.path)
			mtype  = 'image/png'
		elseif  string.find(query.path, '\.ico') then
			mtype  = 'image/jpg'
		elseif  string.find(query.path, '\.js.gz') then
			mtype  = 'application/x-javascript'
		else 
			mtype  = 'text/html'
		end
		
		libhttpd.write( client, "HTTP/1.0 200 OK\r\n" );

		-- if query.path == '/auth' then
		c, v = auth_cookie(parameters)
		if c ~= nil then
			libhttpd.write( client, "Set-Cookie: " .. c .. "=" .. v .. "\r\n" );
			global_cookie = v
		end
		-- end


		libhttpd.write( client, "Server: lua-httpd \r\n" );
		libhttpd.write( client, "Content-type: " .. mtype  .. "; charset=utf-8;\r\n" );
		-- print(query.path ..  "mtype == ".. mtype )
		if string.find(query.path,'\.gz') ~= nil then
			libhttpd.write( client, "Content-Encoding: gzip \r\n");
			-- print ("-------- * * * * ------------")
		end
		
		if string.find(query.path,'\.gz') == nil and string.find(query.path,'\.html') == nil and mtype == 'text/html' then
			libhttpd.write( client, "Expires: Mon, 26 Jul 1997 05:00:00 GMT\r\n" );
			libhttpd.write( client, "Cache-Control: no-cache, must-revalidate\r\n" );
			libhttpd.write( client, "Pragma:no-cache\r\n" );
		else
			libhttpd.write( client, "Last-Modified: Mon, 3 Jan 2005 05:00:00 GMT\r\n" );
			libhttpd.write( client, "Expires: Mon, 5 Jan 2015 05:00:00 GMT\r\n" );
			--libhttpd.write( client, "Cache-Control: public, must-revalidate\r\n" );
			--libhttpd.write( client, "Pragma: cache\r\n" );
		--	libhttpd.write( client, "Cache-Control: cache\r\n" );
		--	print(query.path .. " : Cached")
		end

		libhttpd.write( client, "Connection: close\r\n\r\n" );
		
		r, v = pcall(m, parameters)

		if r ~= true then
			sendError(client, 501, "Error calling the method" .. v)
		end
	else
		sendError(client, 501, "Error getting the method")
	end

	return "-1","200"
end




--
--  Send the given error message to the client.
--  Return the length of data sent to the client so we know what to log.
--
function sendError( client, status, str )
	message = ""
	message = "HTTP/1.0 " .. status .. " OK\r\n" ;
	message = message .. "Server: lua-httpd \r\n";
	message = message .. "Content-type: text/html\r\n";
	message = message .. "Connection: close\r\n\r\n" ;
	message = message .. "<html><head><title>Error</title></head>" ;
	message = message .. "<body><h1>Error</h1" ;
	message = message .. "<p>" .. str .. "</p></body></html>" ;

	libhttpd.write( client, message );
	return string.len(message) ;
end




--
-- Utility function:  URL encoding function
--
--function urlEncode(str)
--	if (str) then
--		str = string.gsub (str, "\n", "\r\n") 
--		str = string.gsub (str, "([^%w ])",
--		function (c) return string.format ("%%%02X", string.byte(c)) end) 
--		str = string.gsub (str, " ", "+") 
--	end 
--	return str 
--end


--
-- Utility function:  URL decode function
--
function urlDecode(str)
	str = string.gsub (str, "+", " ") 
	str = string.gsub (str, "%%(%x%x)", function(h) return string.char(tonumber(h,16)) end) 
	str = string.gsub (str, "\r\n", "\n") 
	return str 
end


--
-- Log an access request.
--
function logAccess( host, ip, request, status, size, agent, referer, major, minor )
	date = os.date("%m/%b/%Y:%H:%M:%S +0000");

	--
	-- Format the logging line.
	--
	log  = string.format( '%s - - [%s] "GET %s HTTP/%s.%s" %s %s "%s" "%s"', ip, date, request, major, minor, status, size, referer, agent );

	logfile =  host .. "/logs/access.log"

	--
	-- Open logfile for appending.
	--

	-- file = io.stdout --io.open( logfile , "a" );
	if ( file ~= nil ) then
		file:write( log .. "\n" );
		--file:close();
	else
		print( "WARNING : Unable to open logfile for writing : " .. logfile );
	end

	--print( log );
end

while 1 do
	local r
	local a
	r, a = pcall(start_server, 8080);
	r, a = pcall(libhttpd.close, listener);
	perax.sleep(1)
	print ('reloading internal server')
end
